/*
 * Copyright (C) 2017 FIX94
 *
 * This software may be modified and distributed under the terms
 * of the MIT license.  See the LICENSE file for details.
 */

#include <stdio.h>
#include <stdbool.h>
#include <inttypes.h>
#include "mapper_h/m1.h"
#include "mapper_h/m4.h"
#include "mapper_h/m4add.h"
#include "mapper_h/m5.h"
#include "mapper_h/m7.h"
#include "mapper_h/m9.h"
#include "mapper_h/m10.h"
#include "mapper_h/m13.h"
#include "mapper_h/m15.h"
#include "mapper_h/m48.h"
#include "mapper_h/m206.h"
#include "mapper_h/m225.h"
#include "mapper_h/m228.h"
#include "mapper_h/namco.h"
#include "mapper_h/p8c8.h"
#include "mapper_h/p16c4.h"
#include "mapper_h/p16c8.h"
#include "mapper_h/p32c4.h"
#include "mapper_h/p32c8.h"
#include "mapper_h/vrc1.h"
#include "mapper_h/vrc2_4.h"
#include "mapper_h/vrc6.h"
#include "mapper_h/vrc7.h"
#include "mapper.h"
#include "mapperList.h"

mapperList_t mapperList[256] = {
	{ p32c8init,	p32c8get8,		m0_set8,		p32c8chrGet8,	p32c8chrSet8,	NULL,	NULL,	NULL },
	{ m1init,		m1get8,			m1set8,			m1chrGet8,		m1chrSet8,		NULL,	NULL,	NULL },
	{ p16c8init,	p16c8get8,		m2_set8,		p16c8chrGet8,	p16c8chrSet8,	NULL,	NULL,	NULL },
	{ p8c8init,		p8c8get8,		m3_set8,		p8c8chrGet8,	p8c8chrSet8,	NULL,	NULL,	NULL },
	{ m4init,		m4get8,			m4set8,			m4chrGet8,		m4chrSet8,		NULL,	NULL,	m4cycle },
	{ m5init,		m5get8,			m5set8,			m5chrGet8,		m5chrSet8,		m5vramGet8, m5vramSet8, m5cycle },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ m7init,		m7get8,			m7set8,			m7chrGet8,		m7chrSet8,		NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ m9init,		m9get8,			m9set8,			m9chrGet8,		m9chrSet8,		NULL,	NULL,	NULL },
	{ m10init,		m10get8,		m10set8,		m10chrGet8,		m10chrSet8,		NULL,	NULL,	NULL },
	{ p32c8init,	p32c8get8,		m11_set8,		p32c8chrGet8,	p32c8chrSet8,	NULL,	NULL,	NULL },
	{ m12init,		m4get8,			m12set8,		m12chrGet8,		m4chrSet8,		NULL,	NULL,	m4cycle },
	{ m13init,		m13get8,		m13set8,		m13chrGet8,		m13chrSet8,		NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ m15init,		m15get8,		m15set8,		m15chrGet8,		m15chrSet8,		NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ namco_init,	namco_get8,		namco_set8,		namco_chrGet8,	namco_chrSet8,	namco_vramGet8,	namco_vramSet8,	namco_cycle },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ vrc2_4_init,	vrc2_4_get8,	m21_set8,		vrc2_4_chrGet8,	vrc2_4_chrSet8,	NULL,	NULL,	vrc2_4_cycle },
	{ vrc2_4_init,	vrc2_4_get8,	m22_set8,		m22_chrGet8,	vrc2_4_chrSet8,	NULL,	NULL,	vrc2_4_cycle },
	{ vrc2_4_init,	vrc2_4_get8,	m23_set8,		vrc2_4_chrGet8,	vrc2_4_chrSet8,	NULL,	NULL,	vrc2_4_cycle },
	{ vrc6init,		vrc6get8,		m24_set8,		vrc6chrGet8,	vrc6chrSet8,	NULL,	NULL,	vrc6cycle },
	{ vrc2_4_init,	vrc2_4_get8,	m25_set8,		vrc2_4_chrGet8,	vrc2_4_chrSet8,	NULL,	NULL,	vrc2_4_cycle },
	{ vrc6init,		vrc6get8,		m26_set8,		vrc6chrGet8,	vrc6chrSet8,	NULL,	NULL,	vrc6cycle },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ m48init,		m48get8,		m33set8,		m48chrGet8,		m48chrSet8,		NULL,	NULL,	NULL },
	{ p32c4init,	p32c4get8,		p32c4set8,		p32c4chrGet8,	p32c4chrSet8,	NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ p32c8init,	m36_p32c8get8,	m36_set8,		p32c8chrGet8,	p32c8chrSet8,	NULL,	NULL,	NULL },
	{ m37_init,		m4get8,			m37_set8,		m4chrGet8,		m4chrSet8,		NULL,	NULL,	m4cycle },
	{ p32c8init,	p32c8get8,		m38_set8,		p32c8chrGet8,	p32c8chrSet8,	NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ p32c8init,	p32c8get8,		m41_set8,		p32c8chrGet8,	p32c8chrSet8,	NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ m44_init,		m4get8,			m44_set8,		m4chrGet8,		m4chrSet8,		NULL,	NULL,	m4cycle },
	{ m45_init,		m4get8,			m45_set8,		m4chrGet8,		m4chrSet8,		NULL,	NULL,	m4cycle },
	{ p32c8init,	p32c8get8,		m46_set8,		p32c8chrGet8,	p32c8chrSet8,	NULL,	NULL,	NULL },
	{ m47_init,		m4get8,			m47_set8,		m4chrGet8,		m4chrSet8,		NULL,	NULL,	m4cycle },
	{ m48init,		m48get8,		m48set8,		m48chrGet8,		m48chrSet8,		NULL,	NULL,	m48cycle },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ m52_init,		m4get8,			m52_set8,		m4chrGet8,		m4chrSet8,		NULL,	NULL,	m4cycle },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ m174_init,	p1632c8get8,	m57_set8,		p16c8chrGet8,	p16c8chrSet8,	NULL,	NULL,	NULL },
	{ p16c8init,	p1632c8get8,	m58_set8,		p16c8chrGet8,	p16c8chrSet8,	NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ p16c8init,	p1632c8get8,	m61_set8,		p16c8chrGet8,	p16c8chrSet8,	NULL,	NULL,	NULL },
	{ p16c8init,	p1632c8get8,	m62_set8,		p16c8chrGet8,	p16c8chrSet8,	NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ p32c8init,	p32c8get8,		m66_set8,		p32c8chrGet8,	p32c8chrSet8,	NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ p16c8init,	p16c8get8,		m70_set8,		p16c8chrGet8,	p16c8chrSet8,	NULL,	NULL,	NULL },
	{ p16c8init,	p16c8get8,		m71_set8,		p16c8chrGet8,	p16c8chrSet8,	NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ vrc1init,		vrc1get8,		vrc1set8,		vrc1chrGet8,	vrc1chrSet8,	NULL,	NULL,	NULL },
	{ m206init,		m206get8,		m206set8,		m76chrGet8,		m206chrSet8,	NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ p16c8init,	p16c8get8,		m78b_set8,		p16c8chrGet8,	p16c8chrSet8,	NULL,	NULL,	NULL },
	{ p32c8init,	p32c8get8,		m79_set8,		p32c8chrGet8,	p32c8chrSet8,	NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ vrc7init,		vrc7get8,		vrc7set8,		vrc7chrGet8,	vrc7chrSet8,	NULL,	NULL,	vrc7cycle },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ p8c8init,		p8c8get8,		m87_set8,		p8c8chrGet8,	p8c8chrSet8,	NULL,	NULL,	NULL },
	{ m206init,		m206get8,		m206set8,		m88chrGet8,		m206chrSet8,	NULL,	NULL,	NULL },
	{ p16c8init,	p16c8get8,		m89_set8,		p16c8chrGet8,	p16c8chrSet8,	NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ p16c8init,	p16c8get8,		m93_set8,		p16c8chrGet8,	p16c8chrSet8,	NULL,	NULL,	NULL },
	{ p16c8init,	p16c8get8,		m94_set8,		p16c8chrGet8,	p16c8chrSet8,	NULL,	NULL,	NULL },
	{ m206init,		m206get8,		m95set8,		m206chrGet8,	m206chrSet8,	NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ p16c8init,	m97_get8,		m97_set8,		p16c8chrGet8,	p16c8chrSet8,	NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ p8c8init,		p8c8get8,		m99_set8,		p8c8chrGet8,	p8c8chrSet8,	NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ p8c8init,		p8c8get8,		m101_set8,		p8c8chrGet8,	p8c8chrSet8,	NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ m206init,		m112get8,		m112set8,		m206chrGet8,	m206chrSet8,	NULL,	NULL,	NULL },
	{ p32c8init,	p32c8get8,		m113_set8,		p32c8chrGet8,	p32c8chrSet8,	NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ m118init,		m4get8,			m118set8,		m4chrGet8,		m4chrSet8,		NULL,	NULL,	m4cycle },
	{ m119init,		m4get8,			m4set8,			m119chrGet8,	m119chrSet8,	NULL,	NULL,	m4cycle },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ p32c8init,	p32c8get8,		m133_set8,		p32c8chrGet8,	p32c8chrSet8,	NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ p32c8init,	p32c8get8,		m140_set8,		p32c8chrGet8,	p32c8chrSet8,	NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ p32c8init,	p32c8get8,		m144_set8,		p32c8chrGet8,	p32c8chrSet8,	NULL,	NULL,	NULL },
	{ p8c8init,		p8c8get8,		m145_set8,		p8c8chrGet8,	p8c8chrSet8,	NULL,	NULL,	NULL },
	{ p32c8init,	p32c8get8,		m79_set8,		p32c8chrGet8,	p32c8chrSet8,	NULL,	NULL,	NULL },
	{ p32c8init,	p32c8get8,		m147_set8,		p32c8chrGet8,	p32c8chrSet8,	NULL,	NULL,	NULL },
	{ p32c8init,	p32c8get8,		m148_set8,		p32c8chrGet8,	p32c8chrSet8,	NULL,	NULL,	NULL },
	{ p8c8init,		p8c8get8,		m149_set8,		p8c8chrGet8,	p8c8chrSet8,	NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ p16c8init,	p16c8get8,		m152_set8,		p16c8chrGet8,	p16c8chrSet8,	NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ m206init,		m206get8,		m154set8,		m88chrGet8,		m206chrSet8,	NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ m174_init,	p1632c8get8,	m174_set8,		p16c8chrGet8,	p16c8chrSet8,	NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ p16c8init,	m180_get8,		m180_set8,		p16c8chrGet8,	p16c8chrSet8,	NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ p16c4init,	p16c4get8,		m184_set8,		p16c4chrGet8,	p16c4chrSet8,	NULL,	NULL,	NULL },
	{ p8c8init,		p8c8get8,		m185_set8,		m185_chrGet8,	p8c8chrSet8,	NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ p16c8init,	m200_get8,		m200_set8,		p16c8chrGet8,	p16c8chrSet8,	NULL,	NULL,	NULL },
	{ p32c8init,	p32c8get8,		m201_set8,		p32c8chrGet8,	p32c8chrSet8,	NULL,	NULL,	NULL },
	{ m174_init,	p1632c8get8,	m202_set8,		p16c8chrGet8,	p16c8chrSet8,	NULL,	NULL,	NULL },
	{ p16c8init,	m200_get8,		m203_set8,		p16c8chrGet8,	p16c8chrSet8,	NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ m205_init,	m4get8,			m205_set8,		m4chrGet8,		m4chrSet8,		NULL,	NULL,	m4cycle },
	{ m206init,		m206get8,		m206set8,		m206chrGet8,	m206chrSet8,	NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ namco_init,	namco_get8,		namco_set8,		namco_chrGet8,	namco_chrSet8,	namco_vramGet8,	namco_vramSet8,	namco_cycle },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ m174_init,	p1632c8get8,	m212_set8,		p16c8chrGet8,	p16c8chrSet8,	NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ m225init,		m225get8,		m225set8,		m225chrGet8,	m225chrSet8,	NULL,	NULL,	NULL },
	{ p16c8init,	p1632c8get8,	m226_set8,		p16c8chrGet8,	p16c8chrSet8,	NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ m228init,		m228get8,		m228set8,		m228chrGet8,	m228chrSet8,	NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ p16c8init,	m231_get8,		m231_set8,		p16c8chrGet8,	p16c8chrSet8,	NULL,	NULL,	NULL },
	{ p16c8init,	m232_get8,		m232_set8,		p16c8chrGet8,	p16c8chrSet8,	NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ p32c8init,	p32c8get8,		m240_set8,		p32c8chrGet8,	p32c8chrSet8,	NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ p32c8init,	p32c8get8,		m242_set8,		p32c8chrGet8,	p32c8chrSet8,	NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
	{ NULL,			NULL,			NULL,			NULL,			NULL,			NULL,	NULL,	NULL },
};
